(*
 This file manages information about position of a token.

 author: Christian Mueller
*)

local 
    open AbstractSyntax Support

    val line_counter = ref 1
    val current_file = ref "none"		   
in

(* info about line number in a file *)

fun getLineno () = !line_counter

fun incLineno () = line_counter := !line_counter + 1

fun resetLineno () = line_counter := 0

(* info about the current file *)

fun getCurrentFile () = !current_file

fun setCurrentFile s = current_file := s

(* info of the parser *)

(* 
   info field is defined in AbstractSyntax due to circular dependendcies 

   type info = line number * start character * file name * identifier option
*)

val emptyInfo = (~1,~1,"toplevel",NONE)

fun hasId (_,_,_,SOME _) = true
  | hasId _ = false

fun setId id (l,c,s,_) = (l,c,s,SOME id)

fun getId (_,_,_,id) = valOf id (* throws option *)

(* sets info according to a function f *)
fun setInfo f (Star info)              = Star (f info)
  | setInfo f (Box info)               = Box (f info)
  | setInfo f (Var (info,v,x))         = Var (f info,v,x)
  | setInfo f (App (info,e1,e2))       = App(f info,e1,e2)
  | setInfo f (Abstr (info,v,e1,e2))   = Abstr(f info,v,e1,e2)
  | setInfo f (Pi (info,v,e1,e2))      = Pi(f info,v,e1,e2) 
  | setInfo f (Int (info,i))           = Int(f info,i)
  | setInfo f (Plus(info,e1,e2))       = Plus(f info,e1,e2)
  | setInfo f (Minus (info,e1,e2))     = Minus(f info,e1,e2)
  | setInfo f (Mult (info,e1,e2))      = Mult(f info,e1,e2)
  | setInfo f (IntNeg (info,e))        = IntNeg(f info,e)
  | setInfo f (IntType info )          = IntType (f info)
  | setInfo f (BoolType info )         = BoolType (f info)
  | setInfo f (IntEqual (info,e1,e2))  = IntEqual(f info,e1,e2)
  | setInfo f (IntNotEqual (info,e1,e2)) = IntNotEqual(f info,e1,e2)
  | setInfo f (Less (info,e1,e2))      = Less(f info,e1,e2)
  | setInfo f (LessEq (info,e1,e2))    = LessEq(f info,e1,e2)
  | setInfo f (Greater (info,e1,e2))   = Greater(f info,e1,e2)
  | setInfo f (GreaterEq (info,e1,e2)) = GreaterEq(f info,e1,e2)
  | setInfo f (True info)              = True (f info)
  | setInfo f (False info)             = False (f info)
  | setInfo f (BoolNeg (info,e))       = BoolNeg(f info,e)
  | setInfo f (If(info,e1,e2,e3))      = If(f info,e1,e2,e3) 
  | setInfo f (UnitType info)          = UnitType (f info)
  | setInfo f (Unit info)              = Unit (f info)
  | setInfo f (VariantSort (info,xs))  = VariantSort(f info,xs)
  | setInfo f (Variant(info,l,e1,e2))  = Variant(f info,l,e1,e2)
  | setInfo f (Case(info,e1,xs))       = Case(f info,e1,xs)
  | setInfo f (RecordSort (info,xs))   = RecordSort(f info,xs)
  | setInfo f (Record (info,xs))       = Record(f info,xs)
  | setInfo f (Proj (info,e,l))        = Proj(f info,e,l)
  | setInfo f (RecursiveSort(info,v,e)) = RecursiveSort(f info,v,e)
  | setInfo f (Fold(info,e1,e2))       = Fold(f info,e1,e2)
  | setInfo f (Unfold(info,e1,e2))     = Unfold(f info,e1,e2)
  | setInfo f (Ascribe(info,e1,e2))    = Ascribe(f info,e1,e2)
  | setInfo f (Fix(info,e))            = Fix(f info, e)
  | setInfo f (Ref(info,e))            = Ref (f info,e)
  | setInfo f (RefType(info,e))        = RefType(f info,e)
  | setInfo f (Deref(info,e))          = Deref(f info,e)
  | setInfo f (Assign(info,e1,e2))     = Assign(f info,e1,e2)
  | setInfo f (Loc(info,l))            = Loc(f info,l)
  | setInfo f (Let(info,id,e1,e2))     = Let(f info,id,e1,e2)

(* sets the identifier in the info field*)
fun setIdentifier s e = setInfo (fn info => setId s info) e

(* clears the identifier field *)
fun clearIdentifier e = setInfo (fn (l,c,f,_) => (l,c,f,NONE)) e

(* get info field *)
fun getInfo (Star info)              = info 
  | getInfo (Box info)               = info 
  | getInfo (Var (info,v,x))         = info 
  | getInfo (App (info,e1,e2))       = info 
  | getInfo (Abstr (info,v,e1,e2))   = info 
  | getInfo (Pi (info,v,e1,e2))      = info  
  | getInfo (Int (info,i))           = info 
  | getInfo (Plus(info,e1,e2))       = info 
  | getInfo (Minus (info,e1,e2))     = info 
  | getInfo (Mult (info,e1,e2))      = info 
  | getInfo (IntNeg (info,e))        = info 
  | getInfo (IntType info )          = info 
  | getInfo (BoolType info )         = info 
  | getInfo (IntEqual (info,e1,e2))  = info 
  | getInfo (IntNotEqual (info,e1,e2)) = info 
  | getInfo (Less (info,e1,e2))      = info 
  | getInfo (LessEq (info,e1,e2))    = info 
  | getInfo (Greater (info,e1,e2))   = info 
  | getInfo (GreaterEq (info,e1,e2)) = info 
  | getInfo (True info)              = info 
  | getInfo (False info)             = info 
  | getInfo (BoolNeg (info,e))       = info 
  | getInfo (If(info,e1,e2,e3))      = info 
  | getInfo (UnitType info)          = info 
  | getInfo (Unit info)              = info 
  | getInfo (VariantSort (info,xs))  = info 
  | getInfo (Variant(info,l,e1,e2))  = info 
  | getInfo (Case(info,e1,xs))       = info 
  | getInfo (RecordSort (info,xs))   = info 
  | getInfo (Record (info,xs))       = info 
  | getInfo (Proj (info,e,l))        = info 
  | getInfo (RecursiveSort(info,v,e)) = info
  | getInfo (Fold(info,e1,e2))       = info 
  | getInfo (Unfold(info,e1,e2))     = info 
  | getInfo (Ascribe(info,e1,e2))    = info 
  | getInfo (Fix(info,e))            = info 
  | getInfo (Ref(info,e))            = info 
  | getInfo (RefType(info,e))        = info 
  | getInfo (Deref(info,e))          = info 
  | getInfo (Assign(info,e1,e2))     = info  
  | getInfo (Loc(info,l))            = info 
  | getInfo (Let(info,_,_,_))        = info

end