{
(*
 Lambda Cube lexer which tokenizes an input stream
 
 author: Christian Mueller
*)

open Lexing Parser Info

(* lexer exception *)

exception Exit
exception LexerError of string * int * int

fun lexerError lexbuf s =  
     raise LexerError (s, Lexing.getLexemeStart lexbuf, 
		       Lexing.getLexemeEnd lexbuf)

val start = ref 0

fun setLexInfo lexbuf token = 
    token (getLineno (), (* line *)
	   Lexing.getLexemeStart lexbuf - !start, (* start character *)    
	   getCurrentFile (),
	   NONE : (string option)
	   )

fun setLexInfo_withvalue lexbuf value token = 
    token ((getLineno (), (* line *)
	    Lexing.getLexemeStart lexbuf - !start, (* start character *)    
	    getCurrentFile (),
	    NONE : (string option)),
	   value)


(* matching of keywords *)
fun keyword lexbuf = case Lexing.getLexeme lexbuf of
            "mu"     => setLexInfo lexbuf MU
          | "fold"   => setLexInfo lexbuf FOLD
          | "unfold" => setLexInfo lexbuf UNFOLD
          | "pi"     => setLexInfo lexbuf PI
          | "Int"    => setLexInfo lexbuf INT_TYPE 
          | "Bool"   => setLexInfo lexbuf BOOL_TYPE
          | "Unit"   => setLexInfo lexbuf UNIT_TYPE
          | "unit"   => setLexInfo lexbuf UNIT
          | "true"   => setLexInfo lexbuf TRUE
          | "false"  => setLexInfo lexbuf FALSE
          | "if"     => setLexInfo lexbuf IF
          | "then"   => setLexInfo lexbuf THEN
          | "else"   => setLexInfo lexbuf ELSE
          | "not"    => setLexInfo lexbuf NOT 
          | "exit"   => raise Exit
          | "case"   => setLexInfo lexbuf CASE          
          | "of"     => setLexInfo lexbuf OF 
          | "as"     => setLexInfo lexbuf AS
          | "fix"    => setLexInfo lexbuf FIX
          | "ref"    => setLexInfo lexbuf REF
          | "Ref"    => setLexInfo lexbuf REFTYPE
          | "let"    => setLexInfo lexbuf LET
          | "in"     => setLexInfo lexbuf IN
          | _        => commands lexbuf

(* matches commands and identifiers *)
and commands lexbuf = 
    let
       val s = Lexing.getLexeme lexbuf
    in 
        case s of
         "def"      => DEFINE
       | "rdef"     => RDEFINE
       | "rstep"    => RSTEP
       | "equiv"    => EQUIV
       | "protocol" => PROTOCOL
       | "help"     => HELP
       | _          => 
           if hd (String.explode s)  = (#"'") 
              then setLexInfo_withvalue lexbuf s FREETYPEID 	   
           else setLexInfo_withvalue lexbuf s ID
     end

(* converts a string representing a number into the corresponding token *)
fun number lexbuf = 
    setLexInfo_withvalue 
	   lexbuf 
	   (valOf (Int.fromString (Lexing.getLexeme lexbuf)))
	   INT

(* needed for handling of nested comments *)
val comment_depth = ref 0

fun inc r = (r := !r + 1)
fun dec r = (r := !r - 1)

}

let space = ` `+

rule Token = parse
  [` ``\t``\r`] { Token lexbuf }
| `\n` {start := Lexing.getLexemeStart lexbuf + 1;
        incLineno (); 
	Token lexbuf }
| "(*" { comment_depth := 1;
         Comment lexbuf;
         Token lexbuf }
| "*)" { lexerError lexbuf "unmatched comment bracket" }
| `*``*`{ setLexInfo lexbuf MULT }
| `*` { setLexInfo lexbuf STAR }
| `.` { setLexInfo lexbuf DOT }
| `@` { setLexInfo lexbuf AT }
| `:` { setLexInfo lexbuf COLON }
| `:``:` { setLexInfo lexbuf DCOLON }
| `;` { setLexInfo lexbuf SEMI }
| `(` { setLexInfo lexbuf LPAR }
| `)` { setLexInfo lexbuf RPAR }
| `[` { setLexInfo lexbuf LSQUARE }
| `]` { setLexInfo lexbuf RSQUARE }
| `{` { setLexInfo lexbuf LCURLY }
| `}` { setLexInfo lexbuf RCURLY }
| `=` { setLexInfo lexbuf EQUAL }
| `\` { setLexInfo lexbuf LAMBDA }
| `<` { setLexInfo lexbuf LT }
| `<``<` { setLexInfo lexbuf LESS }
| `>` { setLexInfo lexbuf GT }
| `>``>` { setLexInfo lexbuf GREATER }
| `<``>` { setLexInfo lexbuf NEQUAL }
| `>``=` { setLexInfo lexbuf GTEQ }
| `<``=` { setLexInfo lexbuf LTEQ }
| `,` { setLexInfo lexbuf COMMA }
| `-` `>` { setLexInfo lexbuf ARROW }
| `=` `>` { setLexInfo lexbuf DARROW }
| `>``=` `>` { setLexInfo lexbuf GT_DARROW }
| `:``=` { setLexInfo lexbuf ASSIGN }
| `!` { setLexInfo lexbuf DEREF }
| `|` { setLexInfo lexbuf VBAR }    
| `\` `/` { setLexInfo lexbuf FORALL }
| `-` `>` { setLexInfo lexbuf ARROW }
| eof { setLexInfo lexbuf EOF }
| `~` { setLexInfo lexbuf UMINUS }
| `-` { setLexInfo lexbuf MINUS }
| `+` { setLexInfo lexbuf PLUS }
| `?` { QMARK }
| "switch" { SwitchCommand lexbuf }
| "use"{ FileName lexbuf } 
| [`0`-`9`]+ { number lexbuf 	  
		 handle Overflow => 
		 lexerError lexbuf "integer constant is too large"}

| [`a`-`z``A`-`Z``'`][`_``a`-`z``A`-`Z``0`-`9``'`]* 
              { keyword lexbuf }

| _   { lexerError lexbuf "Illigal Symbol in input" }

and SwitchCommand = parse
  [` ``\t``\r``\n`] { SwitchCommand lexbuf }
| "tm" { TmCommand lexbuf }
| "pm" { PmCommand lexbuf }
| "sub"   { SWITCH_SUB true }
| "nosub" { SWITCH_SUB false }
| _ { lexerError lexbuf "wrong switch command; use \"pm\" and \"tm\"" }

and PmCommand = parse
  [` ``\t``\r``\n`] { PmCommand lexbuf }
| "o"           { SWITCH_PRINTER_MODE AbstractSyntax.Opaque }
| "ows"         { SWITCH_PRINTER_MODE AbstractSyntax.OpaqueWithSync }
| "opaque"      { SWITCH_PRINTER_MODE AbstractSyntax.Opaque }
| "t"           { SWITCH_PRINTER_MODE AbstractSyntax.Transparent } 
| "transparent" { SWITCH_PRINTER_MODE AbstractSyntax.Transparent } 
| "terms"       { SWITCH_PRINTER_OUTPUT (AbstractSyntax.Term) } 
| "types"       { SWITCH_PRINTER_OUTPUT (AbstractSyntax.Type) }
| "all"         { SWITCH_PRINTER_OUTPUT (AbstractSyntax.All) }
| _             { lexerError lexbuf "unkown pm command" }

and TmCommand = parse
  [` ``\t``\r``\n`] { TmCommand lexbuf }
| "s" { SWITCH_TC_MODE AbstractSyntax.S }
| "f" { SWITCH_TC_MODE AbstractSyntax.F }
| "fw"{ SWITCH_TC_MODE AbstractSyntax.Fw }
| "cc"{ SWITCH_TC_MODE AbstractSyntax.CC }
| "r" { SWITCH_TC_MODE AbstractSyntax.R }
| "rg"{ SWITCH_TC_MODE AbstractSyntax.RG }
| "u" { SWITCH_TC_MODE AbstractSyntax.U }
| "ud"{ SWITCH_TC_MODE AbstractSyntax.UD }
| _   { lexerError lexbuf "unkown tm command" }

and FileName = parse
  [` ``\t``\r``\n`] { FileName lexbuf }
| `"`[`:``\\``a`-`z``A`-`Z``0`-`9``/``_``.``~`]*`"` 
       { let 
	   val f = tl(String.explode(getLexeme lexbuf))
	   val f' = rev(tl(rev f))
         in 
             USE_FILE (implode f')
         end 
       }
| _   { lexerError lexbuf "wrong filename format; use \"filename\"" }

and Comment = parse
  "(*" { inc comment_depth; Comment lexbuf }
| "*)" { dec comment_depth; 
         if !comment_depth > 0 then Comment lexbuf else () }
| eof  { lexerError lexbuf "comment not terminated" }
| _    { Comment lexbuf }
;

}

